#include <stdint.h>
#include "sdk_common.h"
#include "ble_zive.h"
#include "ble_srv_common.h"
#include "sdk_macros.h"
#include "nrf_log.h"

/**@brief Function for handling the Connect event.
 *
 * @param[in]   p_zive      Zive Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(ble_zive_t * p_zive, ble_evt_t const * p_ble_evt) {
    p_zive->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
}


/**@brief Function for handling the Disconnect event.
 *
 * @param[in]   p_zive      Zive Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_zive_t * p_zive, ble_evt_t const * p_ble_evt) {
    UNUSED_PARAMETER(p_ble_evt);
    p_zive->conn_handle = BLE_CONN_HANDLE_INVALID;
    ble_zive_evt_t evt;
    evt.evt_type = BLE_ZIVE_EVT_DISCONNECT;
    p_zive->evt_handler(p_zive, &evt);
}

/**@brief Function for handling write events to the Zive ECG data Client Characteristic Configuration Descriptor characteristic.
 *
 * @param[in]   p_zive        Zive Service structure.
 * @param[in]   p_evt_write   Write event received from the BLE stack.
 */
static void on_zive_cccd_write(ble_zive_t * p_zive, ble_gatts_evt_write_t const * p_evt_write) {
    if (p_evt_write->len == 2) {
        // CCCD written, update notification state
        if (p_zive->evt_handler != NULL) {
            ble_zive_evt_t evt;

            if (ble_srv_is_notification_enabled(p_evt_write->data)) {
                evt.evt_type = BLE_ZIVE_EVT_NOTIFICATION_ENABLED;
            } else {
                evt.evt_type = BLE_ZIVE_EVT_NOTIFICATION_DISABLED;
            }

            p_zive->evt_handler(p_zive, &evt);
        }
    }
}

static void on_zive_config_write(ble_zive_t * p_zive, ble_gatts_evt_write_t const * p_evt_write) {
	// nothing for now
}

/**@brief Function for handling the Write event.
 *
 * @param[in]   p_zive      Zive Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_write(ble_zive_t * p_zive, ble_evt_t const * p_ble_evt) {
    ble_gatts_evt_write_t const * p_evt_write = &p_ble_evt->evt.gatts_evt.params.write;

    if (p_evt_write->handle == p_zive->data_handles.cccd_handle) {
    	on_zive_cccd_write(p_zive, p_evt_write);
    } else if(p_evt_write->handle == p_zive->ctrl_handles.value_handle) {
    	on_zive_config_write(p_zive, p_evt_write);
    }
}

void ble_zive_on_ble_evt(ble_evt_t const * p_ble_evt, void * p_context) {
    ble_zive_t * p_zive = (ble_zive_t *) p_context;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            on_connect(p_zive, p_ble_evt);
            break;

        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnect(p_zive, p_ble_evt);
            break;

        case BLE_GATTS_EVT_WRITE:
            on_write(p_zive, p_ble_evt);
            break;

        default:
            // No implementation needed.
            break;
    }
}

#define OPCODE_LENGTH 1                                                              /**< Length of opcode inside Heart Rate Measurement packet. */
#define HANDLE_LENGTH 2                                                              /**< Length of handle inside Heart Rate Measurement packet. */
#define MAX_DATA_LEN      (NRF_SDH_BLE_GATT_MAX_MTU_SIZE - OPCODE_LENGTH - HANDLE_LENGTH) /**< Maximum size of a transmitted Heart Rate Measurement. */

uint8_t initial[1];

uint32_t ble_zive_init(ble_zive_t * p_zive, const ble_zive_init_t * p_zive_init)
{
    uint32_t              err_code;
    ble_uuid_t            ble_uuid;
    ble_add_char_params_t add_char_params;
    //uint8_t               encoded_initial_hrm[MAX_HRM_LEN];

    // Initialize service structure
    p_zive->evt_handler                 = p_zive_init->evt_handler;
    p_zive->conn_handle                 = BLE_CONN_HANDLE_INVALID;
    p_zive->is_sensor_contact_detected  = false;
    p_zive->max_data_len                 = MAX_DATA_LEN;

    // Add service.
    ble_uuid128_t base_uuid = {ZIVE_UUID_BASE};
    err_code = sd_ble_uuid_vs_add(&base_uuid, &p_zive->uuid_type);

    if (err_code != NRF_SUCCESS) {
        return err_code;
    }

    ble_uuid.type = p_zive->uuid_type;
    ble_uuid.uuid = ZIVE_UUID_SERVICE;

    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &p_zive->service_handle);

    if (err_code != NRF_SUCCESS) {
        return err_code;
    }

    // Add heart rate measurement characteristic
    memset(&add_char_params, 0, sizeof(add_char_params));

    add_char_params.uuid              = ZIVE_UUID_DATA_CHAR;
    add_char_params.max_len           = MAX_DATA_LEN;
    add_char_params.init_len          = 0;//TODO hrm_encode(p_hrs, INITIAL_VALUE_HRM, encoded_initial_hrm);
    add_char_params.p_init_value      = initial;
    add_char_params.is_var_len        = true;
    add_char_params.char_props.notify = 1;
    add_char_params.cccd_write_access = p_zive_init->data_cccd_wr_sec;

    err_code = characteristic_add(p_zive->service_handle, &add_char_params, &(p_zive->data_handles));
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }

	// Add control characteristic
	memset(&add_char_params, 0, sizeof(add_char_params));

	add_char_params.uuid            = ZIVE_UUID_CONFIG_CHAR;
	add_char_params.max_len         = sizeof(uint8_t);
	add_char_params.init_len        = sizeof(uint8_t);
	add_char_params.p_init_value    = initial;
	add_char_params.char_props.read = 1;
	add_char_params.char_props.write_wo_resp = 1;
	add_char_params.read_access     = p_zive_init->ctrl_rd_sec;

	err_code = characteristic_add(p_zive->service_handle, &add_char_params, &(p_zive->ctrl_handles));
	if (err_code != NRF_SUCCESS) {
		return err_code;
	}

    return NRF_SUCCESS;
}

uint32_t ble_zive_ecg_send(ble_zive_t * p_zive, int32_t * p_data, uint16_t length){
    uint32_t err_code;
    NRF_LOG_INFO("Sending %d bytes", length);
    // Send value if connected and notifying
    if (p_zive->conn_handle != BLE_CONN_HANDLE_INVALID) {

        uint16_t               hvx_len;
        ble_gatts_hvx_params_t hvx_params;

        hvx_len = length;

        memset(&hvx_params, 0, sizeof(hvx_params));

        hvx_params.handle = p_zive->data_handles.value_handle;
        hvx_params.type   = BLE_GATT_HVX_NOTIFICATION;
        hvx_params.offset = 0;
        hvx_params.p_len  = &hvx_len;
        hvx_params.p_data = (uint8_t *) p_data;

        err_code = sd_ble_gatts_hvx(p_zive->conn_handle, &hvx_params);
        if ((err_code == NRF_SUCCESS) && (hvx_len != length)) {
            err_code = NRF_ERROR_DATA_SIZE;
        }

        NRF_LOG_INFO("%d bytes sent", *(hvx_params.p_len));
    } else {
        err_code = NRF_ERROR_INVALID_STATE;
    }

    return err_code;
}

void ble_zive_on_gatt_evt(ble_zive_t * p_zive, nrf_ble_gatt_evt_t const * p_gatt_evt)
{
    if (    (p_zive->conn_handle == p_gatt_evt->conn_handle)
        &&  (p_gatt_evt->evt_id == NRF_BLE_GATT_EVT_ATT_MTU_UPDATED))
    {
    	NRF_LOG_INFO("Setting MTU to %d", p_gatt_evt->params.att_mtu_effective);
        p_zive->max_data_len = p_gatt_evt->params.att_mtu_effective - OPCODE_LENGTH - HANDLE_LENGTH;
    }
}
